/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * The APR LDAP bind functions provide the ability to bind to a directory
 * in a generic way.
 *
 * First we try binding interactively using SASL, falling back to SASL bind,
 * then simple bind, as supported on the platform.
 *
 * @file apr_ldap_bind.h
 * @brief Apache LDAP library
 */

#ifndef APU_LDAP_BIND_H
#define APU_LDAP_BIND_H

/**
 * @addtogroup APR_Util_LDAP
 * @{
 **/

#if defined(DOXYGEN)
#include "apr_ldap.h"
#endif

/*
 * Handle the case when LDAP is enabled
 */
#if APR_HAS_LDAP

/** LDAP interaction identifiers */
typedef enum {
    APR_LDAP_INTERACT_GETREALM = 0x4008,          /**< SASL realm for the authentication attempt */
    APR_LDAP_INTERACT_AUTHNAME = 0x4002,          /**< SASL username to authenticate */
    APR_LDAP_INTERACT_USER = 0x4001,              /**< SASL username to use for proxy authorization */
    APR_LDAP_INTERACT_PASS = 0x4004,              /**< SASL password for the provided username */
    APR_LDAP_INTERACT_NOECHOPROMPT = 0x4006,      /**< SASL generic prompt for input with input echoing disabled */
    APR_LDAP_INTERACT_ECHOPROMPT = 0x4005,        /**< SASL generic prompt for input with input echoing enabled */
} apr_ldap_bind_interact_e;

typedef struct apr_ldap_bind_interact_t {
    apr_ldap_bind_interact_e id; /* same as client/user callback ID */
    const char *challenge;       /* presented to user (e.g. OTP challenge) */
    const char *prompt;          /* presented to user (e.g. "Username: ") */
    const char *defresult;       /* default result string */
    const void *result;          /* set to point to result */
    apr_size_t len;              /* set to length of result */
} apr_ldap_bind_interact_t;

typedef apr_status_t (apr_ldap_interact_proc)(
        apr_ldap_t *ld, unsigned int flags, apr_ldap_bind_interact_t *interact, void *ctx);

typedef struct apr_ldap_message_t apr_ldap_message_t;


/**
 * APR result function.
 *
 * This function returns the result of a previous request, ready for further
 * processing.
 *
 * @return APR_INCOMPLETE means that at least one result has been
 * returned, and needs to be passed back to the function that
 * generated this message for further processing.
 * APR_SUCCESS means that no further messages are available, and
 * results are complete. Other error codes indicate that the message
 * read was not successful.
 */
APU_DECLARE_LDAP(apr_status_t) apr_ldap_result(apr_ldap_message_t *msg,
                                               apr_interval_time_t timeout,
                                               apu_err_t *result)
                                               __attribute__((nonnull(1,3)));

/**
 * APR LDAP bind function
 *
 * This function binds a previously initialised LDAP connection
 * to the directory.
 *
 * Binds are attempted as SASL interactive, falling back to a
 * standard SASL bind, falling back to a simple bind, depending
 * on the capabilities of the platform.
 *
 * Binds are attempted asynchronously. If APR_INCOMPLETE is returned,
 * this function must be called again.
 *
 * @return APR_INCOMPLETE means that processing has occurred, and
 * the message in reply needs to be fetched using apr_ldap_return().
 * APR_SUCCESS means that the processing is complete, and the bind
 * has been successful. Other error codes indicate that the bind
 * was not successful.
 */
APU_DECLARE_LDAP(apr_status_t) apr_ldap_bind(apr_ldap_t *ldap,
                                             const char *dn,
                                             const char *mech,
                                             apr_ldap_interact_proc *interact,
                                             void *ctx,
                                             apr_interval_time_t timeout,
                                             apr_ldap_message_t **msg,
                                             apu_err_t *result)
                                             __attribute__((nonnull(1,4,7,8)));

#endif /* APR_HAS_LDAP */

/** @} */

#endif /* APU_LDAP_BIND_H */

