/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * apr_ldap_bind.c: LDAP v2/v3 common initialise
 *
 */

#include "apr.h"
#include "apu.h"
#include "apu_config.h"

#if APU_DSO_BUILD
#define APU_DSO_LDAP_BUILD
#endif

#include "apr_ldap.h"
#include "apr_ldap_private.h"
#include "apu_internal.h"
#include "apr_errno.h"
#include "apr_pools.h"
#include "apr_strings.h"

#ifdef HAVE_SASL_SASL_H
#include <sasl/sasl.h>
#else
#ifdef HAVE_SASL_H
#include <sasl.h>
#endif
#endif

#if APR_HAS_LDAP

typedef struct apr_ldap_bind_ctx_t {
    apr_ldap_t *ld;
    apr_ldap_interact_proc *interact;
    void *ctx;
    apr_status_t status;
} apr_ldap_bind_ctx_t;

APU_DECLARE_LDAP(apr_status_t) apr_ldap_message_clear(apr_ldap_message_t *msg)
{
    if (msg) {
        msg->rmech = NULL;
        if (msg->message) {
            ldap_msgfree(msg->message);
            msg->message = NULL;
        }
        msg->msgid = 0;
    }

    return APR_SUCCESS;
}

static apr_status_t message_cleanup(void *dptr)
{
    return apr_ldap_message_clear(dptr);
}

APU_DECLARE_LDAP(apr_ldap_message_t *) apr_ldap_message_make(apr_ldap_t *ldap)
{
    apr_ldap_message_t *msg = apr_pcalloc(ldap->pool, sizeof(apr_ldap_message_t));

    if (msg) {
        msg->ld = ldap;
        apr_pool_cleanup_register(ldap->pool, msg, message_cleanup,
                                  apr_pool_cleanup_null);
    }

    return msg;
}

APU_DECLARE_LDAP(apr_status_t) apr_ldap_result(apr_ldap_message_t *msg,
                                               apr_interval_time_t timeout,
                                               apu_err_t *result)
{
    int res;

    LDAP *ld = msg->ld->ld;

    struct timeval tv, *tvptr;

    if (timeout < 0) {
        tvptr = NULL;
    }
    else {
        tv.tv_sec = (long) apr_time_sec(timeout);
        tv.tv_usec = (long) apr_time_usec(timeout);
        tvptr = &tv;
    }

    if (msg->message) {
        ldap_msgfree( msg->message );
    }

    res = ldap_result( ld, msg->msgid, LDAP_MSG_RECEIVED, tvptr, &msg->message );
    if ( res == -1 ) {
        result->reason = "LDAP: ldap_result() retrieval failed";

        /* -1 is LDAP_SERVER_DOWN in openldap, use something else */
#ifdef LDAP_OPT_ERROR_NUMBER
        ldap_get_option(ld, LDAP_OPT_ERROR_NUMBER, &result->rc);
#endif 
#ifdef LDAP_OPT_RESULT_CODE
        ldap_get_option(ld, LDAP_OPT_RESULT_CODE, &result->rc);
#endif 
        result->msg = ldap_err2string(result->rc);
        return apr_ldap_status(result->rc, APR_EGENERAL);
    }
    else if (res == 0) {
        result->reason = "LDAP: ldap_result() timed out";
        result->rc = LDAP_TIMEOUT;
        result->msg = ldap_err2string(result->rc);
        return APR_ETIMEDOUT;
    }

    res = ldap_count_messages(ld, msg->message);
    if ( res == -1 ) {
        result->reason = "LDAP: ldap_count_entries() retrieval failed";

        /* -1 is LDAP_SERVER_DOWN in openldap, use something else */
#ifdef LDAP_OPT_ERROR_NUMBER
        ldap_get_option(ld, LDAP_OPT_ERROR_NUMBER, &result->rc);
#endif
#ifdef LDAP_OPT_RESULT_CODE
        ldap_get_option(ld, LDAP_OPT_RESULT_CODE, &result->rc);
#endif
        result->msg = ldap_err2string(result->rc);
        return apr_ldap_status(result->rc, APR_EGENERAL);
    }
    else if (res == 0) {
        return APR_SUCCESS;
    }
    else {
        return APR_INCOMPLETE;
    }

}


static int bind_sasl_interact(LDAP *ld, unsigned flags, void *ctx, void *in)
{
    apr_ldap_bind_ctx_t *payload = ctx;
    sasl_interact_t *sasl_interact = in;

    if (!ld) {
        return LDAP_PARAM_ERROR;
    }

    while( sasl_interact->id != SASL_CB_LIST_END ) {

        apr_ldap_bind_interact_t interaction;

        apr_status_t status;

        if (!payload->interact) {
            return LDAP_PARAM_ERROR;
        }

        interaction.id = sasl_interact->id;
        interaction.challenge = sasl_interact->challenge;
        interaction.prompt = sasl_interact->prompt;
        interaction.defresult = sasl_interact->defresult;
        interaction.result = NULL;
        interaction.len = 0;

        status = payload->interact(payload->ld, flags, &interaction, payload->ctx);

        if (status != APR_SUCCESS) {
            payload->status = status;
            return LDAP_PARAM_ERROR;
        }

        sasl_interact->result = interaction.result;
        sasl_interact->len = interaction.len;

        sasl_interact++;
    }

    return LDAP_SUCCESS;
}


/**
 * APR LDAP bind function
 *
 * This function binds a previously initialised LDAP connection
 * to the directory.
 *
 * Binds are attempted as SASL interactive, falling back to a
 * standard SASL bind, falling back to a simple bind, depending
 * on the capabilities of the platform.
 *
 * Binds are attempted asynchronously. If APR_EAGAIN is returned,
 * this function must be called again.
 */
APU_DECLARE_LDAP(apr_status_t) apr_ldap_bind(apr_ldap_t *ldap,
                                             const char *dn,
                                             const char *mech,
                                             apr_ldap_interact_proc *interact,
                                             void *ctx,
                                             apr_interval_time_t timeout,
                                             apr_ldap_message_t **msg,
                                             apu_err_t *result)
{

    LDAPControl *sctrls[] = { 0 };
    LDAPControl *cctrls[] = { 0 };

    unsigned int flags = LDAP_SASL_QUIET;

    apr_ldap_bind_ctx_t payload;

    struct timeval tv, *tvptr;

    if (timeout < 0) {
        tvptr = NULL;
    }
    else {
        tv.tv_sec = (long) apr_time_sec(timeout);
        tv.tv_usec = (long) apr_time_usec(timeout);
        tvptr = &tv;
    }

    memset(result, 0, sizeof(*result));

    payload.ld = ldap;
    payload.interact = interact;
    payload.ctx = ctx;
    payload.status = APR_SUCCESS;

    if (!*msg) {
        *msg = apr_ldap_message_make(ldap);
        if (!*msg) {
            return APR_ENOMEM;
        }
    }

#ifdef LDAP_OPT_NETWORK_TIMEOUT
    if (!(*msg)->message) {
        result->rc = ldap_set_option(ldap->ld, LDAP_OPT_NETWORK_TIMEOUT, tvptr);
        if (result->rc != LDAP_SUCCESS) {
            result->msg = ldap_err2string(result->rc);
            result->reason = "LDAP: Could not set network timeout";
            return APR_EINVAL;
        }
    }
#endif

    if (dn) {

        /* Setting a distinguished name means we want a simple bind */

        if (!(*msg)->msgid) {

            apr_ldap_bind_interact_t interaction = { 0 };

            interaction.id = APR_LDAP_INTERACT_PASS;
            interaction.prompt = "Password";

            payload.status = interact(ldap, 0, &interaction, ctx);

            if (payload.status != APR_SUCCESS) {
                return payload.status;
            }

            (*msg)->msgid = ldap_simple_bind(ldap->ld, dn, interaction.result);

            if ((*msg)->msgid > -1) {
                result->rc = LDAP_SUCCESS;
                return APR_INCOMPLETE;
            }
            else {

#ifdef LDAP_OPT_ERROR_NUMBER
                ldap_get_option(ldap->ld, LDAP_OPT_ERROR_NUMBER, &result->rc);
#endif
#ifdef LDAP_OPT_RESULT_CODE
                ldap_get_option(ldap->ld, LDAP_OPT_RESULT_CODE, &result->rc);
#endif

                apr_ldap_message_clear(*msg);
                result->msg = ldap_err2string(result->rc);
                result->reason = "LDAP: ldap_simple_bind() failed";
                return apr_ldap_status(result->rc, APR_EGENERAL);
            }

        } 
        else {

            int res = ldap_parse_result(ldap->ld, (*msg)->message, &result->rc, NULL, NULL, NULL,
                                 NULL, 0);

            apr_ldap_message_clear(*msg);

            if (res > -1) {
                return APR_SUCCESS;
            }
            else {
                result->msg = ldap_err2string(result->rc);
                result->reason = "LDAP: ldap_parse_result() failed";
                return apr_ldap_status(result->rc, APR_EGENERAL);
            }

        }

    }
    else {

#if APR_HAS_LDAP_SASL_INTERACTIVE_BIND

        /* No distinguished name is a SASL bind */

        result->rc = ldap_sasl_interactive_bind(ldap->ld, dn, mech,
                                                sctrls, cctrls, flags, bind_sasl_interact, &payload,
                                                (*msg)->message, &(*msg)->rmech, &(*msg)->msgid );

        if (APR_SUCCESS != payload.status) {
            apr_ldap_message_clear(*msg);
            return payload.status;
        }
        else if (result->rc == LDAP_SUCCESS) {
            apr_ldap_message_clear(*msg);
            return APR_SUCCESS;
        }
        else if (result->rc == LDAP_SASL_BIND_IN_PROGRESS) {
            return APR_INCOMPLETE;
        }
        else {
            apr_ldap_message_clear(*msg);
            result->msg = ldap_err2string(result->rc);
            result->reason = "LDAP: ldap_sasl_interactive_bind() failed";
            return apr_ldap_status(result->rc, APR_EGENERAL);
        }

#else

        /*
         * for platforms that do not support ldap_sasl_interactive_bind(), alternative
         * implementations using ldap_sasl_bind() go here.
         */

        result->reason = "LDAP: SASL bind not yet supported by APR on this "
                         "LDAP SDK";
        result->rc = LDAP_UNWILLING_TO_PERFORM;
        return APR_ENOTIMPL;
#endif

    }

}

#endif /* APR_HAS_LDAP */

